#ifndef _cover_hpp_INCLUDED
#define _cover_hpp_INCLUDED

/*------------------------------------------------------------------------*/

// This file has dual usage: 'COVER' macro for testing and debugging but
// also contains data structures for 'covered clause elimination' which is
// also 'cover' for short.  That is why we merged those two files.

/*------------------------------------------------------------------------*/

// Coverage goal, used similar to 'assert' (but with flipped condition) and
// also included even if 'NDEBUG' is defined (in optimizing compilation).
//
// This should in essence not be used in production code.
//
// There seems to be no problem overloading the name 'COVER' of this macro
// with the constant 'COVER' of 'Internal::Mode' (surprisingly).

#define COVER(COND) \
do { \
  if (!(COND)) break; \
  fprintf (stderr, \
    "%scadical%s: %s:%d: %s: Coverage goal %s`%s'%s reached.\n", \
    terr.bold_code (), terr.normal_code (), \
    __FUNCTION__, __LINE__, __FILE__, \
    terr.green_code (), # COND, terr.normal_code ()); \
  fflush (stderr); \
  abort (); \
} while (0)

/*------------------------------------------------------------------------*/

namespace CaDiCaL {

struct Coveror {
  std::vector<int> added;
  std::vector<int> covered;
  struct { size_t asymmetric, covered; } next;
  Coveror () { next.asymmetric = next.covered = 0; }
};

}

#endif
